/*
    This file is part of Mitsuba, a physically based rendering system.

    Copyright (c) 2007-2011 by Wenzel Jakob and others.

    Mitsuba is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3
    as published by the Free Software Foundation.

    Mitsuba is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <mitsuba/render/phase.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/core/frame.h>
#include <mitsuba/core/pdf.h>
#include <boost/math/special_functions.hpp>
#include <fstream>

MTS_NAMESPACE_BEGIN

/**
 * Phase function read from file
 */
class FilePhaseFunction : public PhaseFunction {
public:
	FilePhaseFunction(const Properties &props)
		: PhaseFunction(props),
		  m_step(0) {
		m_file = props.getString("file");
	}

	FilePhaseFunction(Stream *stream, InstanceManager *manager)
		: PhaseFunction(stream, manager) {
		m_file = stream->readString();
		configure();
	}

	virtual ~FilePhaseFunction() { }


	void serialize(Stream *stream, InstanceManager *manager) const {
		PhaseFunction::serialize(stream, manager);
		stream->writeString(m_file);
	}

	void configure() {
		std::ifstream inFile(m_file.c_str());
		size_t numSteps;
		if (!inFile.good()) {
			Log(EError, "Cannot read input file: (numSteps, step).");
		}
		inFile >> numSteps >> m_step;
		m_pdfsampled.resize(numSteps, 0);
		m_pdf = TabulatedPDF(numSteps);
		if (!inFile.good()) {
			Log(EError, "Cannot read input file: entry " SIZE_T_FMT " entry.", \
					0);
		}
		inFile >> m_pdfsampled[0];
		Float theta = 0;
		m_pdf[0] = 0.0f;
		for (size_t iterX = 1; iterX < numSteps; ++iterX) {
			theta += m_step;
			if (!inFile.good()) {
				Log(EError, "Cannot read input file: entry " SIZE_T_FMT " entry.", \
						iterX);
			}
			inFile >> m_pdfsampled[iterX];
			m_pdf[iterX] = m_pdfsampled[iterX] * 2.0 * M_PI * std::sin(theta);
			m_pdf.setRVEntry(iterX, theta);
		}
		inFile.close();
		m_pdf.build();
		Log(EInfo, toString().c_str());
	}

	inline Float sample(PhaseFunctionQueryRecord &pRec,
			Sampler *sampler) const {
		Point2 sample(sampler->next2D());

		Float theta = m_pdf.sample(sample.x);
		Float cosTheta = std::cos(theta);
		Float sinTheta = std::sqrt(std::max((Float) 0, 1.0f-cosTheta*cosTheta));
		Float phi = 2*M_PI*sample.y;
		Float cosPhi = std::cos(phi);
		Float sinPhi = std::sin(phi);

		Vector dir(
			sinTheta * cosPhi,
			sinTheta * sinPhi,
			cosTheta);
		pRec.wo = Frame(-pRec.wi).toWorld(dir);
		return 1.0f;
	}

	Float sample(PhaseFunctionQueryRecord &pRec,
			Float &pdf, Sampler *sampler) const {
		FilePhaseFunction::sample(pRec, sampler);
		pdf = FilePhaseFunction::eval(pRec);
		return 1.0f;
	}

	Float eval(const PhaseFunctionQueryRecord &pRec) const {
		Float theta = unitAngle(-pRec.wi, pRec.wo);
		if (theta < 0) {
			theta = 2.0 * M_PI + theta;
		}
		if (theta > M_PI) {
			theta = 2.0 * M_PI - theta;
		}
		return m_pdfsampled[static_cast<int>(std::floor(theta / m_step + 0.5))];
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << "FilePhaseFunction[" << endl
			<< "   file 		= " << m_file << endl
			<< "   step 		= " << m_step << endl;
//			for(size_t iter = 0; iter < m_pdf.size(); ++iter) {
//				oss << "iter " << iter \
//					<< " theta " << m_pdf.getRVEntry(iter) \
//					<< " pdfs " << m_pdfsampled[iter] \
//					<< " pdf " << m_pdf[iter] \
//					<< " cdf " << m_pdf.getCDFEntry(iter)
//					<< endl;
//			}
			oss << "]";
		return oss.str();
	}

	MTS_DECLARE_CLASS()

private:
	std::string m_file;
    TabulatedPDF m_pdf;
    std::vector<Float> m_pdfsampled;
	Float m_step;
};


MTS_IMPLEMENT_CLASS_S(FilePhaseFunction, false, PhaseFunction)
MTS_EXPORT_PLUGIN(FilePhaseFunction, "Phase function read from file");
MTS_NAMESPACE_END
